
/* Common Flash Interface structures 
 * See http://support.intel.com/design/flash/technote/index.htm
 * $Id: cfi.h,v 1.3 2000/09/27 17:56:00 jamey Exp $
 */

#ifndef __MTD_CFI_H__
#define __MTD_CFI_H__

#include <linux/types.h>
#include <linux/mtd/flashchip.h>

/*
 * You can optimize the code size and performance by defining only 
 * the geometry(ies) available on your hardware.
 * CFIDEV_INTERLEAVE_n, where  represents the interleave (number of chips to fill the bus width)
 * CFIDEV_BUSWIDTH_n, where n is the bus width in bytes (1, 2 or 4 bytes)
 *
 * By default, all (known) geometries are supported.
 */
#define CFIDEV_INTERLEAVE_1 (1)
#define CFIDEV_INTERLEAVE_2 (2)
#define CFIDEV_INTERLEAVE_4 (4)

#if !defined(CFIDEV_INTERLEAVE_1) && !defined(CFIDEV_INTERLEAVE_2) && !defined(CFIDEV_INTERLEAVE_4)
#error You must define at least one interleave to support!
#endif

#define CFIDEV_BUSWIDTH_1 (1)
#define CFIDEV_BUSWIDTH_2 (2)
#define CFIDEV_BUSWIDTH_4 (4)

#if !defined(CFIDEV_BUSWIDTH_1) && !defined(CFIDEV_BUSWIDTH_2) && !defined(CFIDEV_BUSWIDTH_4)
#error You must define at least one bus width to support!
#endif

/* NB: these values must represents the number of bytes needed to meet the 
 *     device type (x8, x16, x32).  Eg. a 32 bit device is 4 x 8 bytes. 
 *     These numbers are used in calculations.
 */
#define CFI_DEVICETYPE_X8  (8 / 8)
#define CFI_DEVICETYPE_X16 (16 / 8)
#define CFI_DEVICETYPE_X32 (32 / 8)

/* NB: We keep these structures in memory in HOST byteorder, except
 * where individually noted.
 */

/* Basic Query Structure */
struct cfi_ident {
  __u8  qry[3];
  __u16 P_ID;
  __u16 P_ADR;
  __u16 A_ID;
  __u16 A_ADR;
  __u8  VccMin;
  __u8  VccMax;
  __u8  VppMin;
  __u8  VppMax;
  __u8  WordWriteTimeoutTyp;
  __u8  BufWriteTimeoutTyp;
  __u8  BlockEraseTimeoutTyp;
  __u8  ChipEraseTimeoutTyp;
  __u8  WordWriteTimeoutMax;
  __u8  BufWriteTimeoutMax;
  __u8  BlockEraseTimeoutMax;
  __u8  ChipEraseTimeoutMax;
  __u8  DevSize;
  __u16 InterfaceDesc;
  __u16 MaxBufWriteSize;
  __u8  NumEraseRegions;
  __u32 EraseRegionInfo[1]; /* Not host ordered */
} __attribute__((packed));

/* Extended Query Structure for both PRI and ALT */

struct cfi_extquery {
  __u8  pri[3];
  __u8  MajorVersion;
  __u8  MinorVersion;
} __attribute__((packed));

/* Vendor-Specific PRI for Intel/Sharp Extended Command Set (0x0001) */

struct cfi_pri_intelext {
  __u8  pri[3];
  __u8  MajorVersion;
  __u8  MinorVersion;
  __u32 FeatureSupport;
  __u8  SuspendCmdSupport;
  __u16 BlkStatusRegMask;
  __u8  VccOptimal;
  __u8  VppOptimal;
} __attribute__((packed));

struct cfi_pri_query {
  __u8  NumFields;
  __u32 ProtField[1]; /* Not host ordered */
} __attribute__((packed));

struct cfi_bri_query {
  __u8  PageModeReadCap;
  __u8  NumFields;
  __u32 ConfField[1]; /* Not host ordered */
} __attribute__((packed));

#define P_ID_NONE 0
#define P_ID_INTEL_EXT 1
#define P_ID_AMD_STD 2
#define P_ID_INTEL_STD 3
#define P_ID_AMD_EXT 4
#define P_ID_MITSUBISHI_STD 256
#define P_ID_MITSUBISHI_EXT 257
#define P_ID_RESERVED 65535


struct cfi_private {
	__u16 cmdset;
	void *cmdset_priv;
	int interleave;
	int device_type;
	struct mtd_info *(*cmdset_setup)(struct map_info *);
	struct cfi_ident cfiq; /* For now only one. We insist that all devs
				  must be of the same type. */
	int numchips;
	unsigned long chipshift; /* Because they're of the same type */
	struct flchip chips[0];  /* per-chip data structure for each chip */
};

#define MAX_CFI_CHIPS 8 /* Entirely arbitrary to avoid realloc() */

/*
 * Returns the command address according to the given geometry.
 */
extern __inline__ __u32 cfi_build_cmd_addr(__u32 cmd_ofs, u_char interleave, int type)
{
	return (cmd_ofs * type) * interleave;
}

/*
 * Transforms the CFI command for the given geometry (bus width & interleave.
 */
extern __inline__ __u32 cfi_build_cmd(u_char cmd, struct map_info *map, u_char interleave)
{
	__u32 val = 0;

	switch(map->buswidth) {
#ifdef CFIDEV_BUSWIDTH_1
	case CFIDEV_BUSWIDTH_1: /* 1 x8 device */
		val = cmd;
		break;
#endif
#ifdef CFIDEV_BUSWIDTH_2
	case CFIDEV_BUSWIDTH_2:
		switch (interleave) {
#ifdef CFIDEV_INTERLEAVE_1
		case CFIDEV_INTERLEAVE_1: /* 1 x16 device in x16 mode */
			val = cmd;
			break;
#endif
#ifdef CFIDEV_INTERLEAVE_2
		case CFIDEV_INTERLEAVE_2: /* 2 (x8, x16 or x32) devices in x8 mode */
			val = (cmd << 8) | cmd;
			break;
		}
#endif
		break;
#endif /* CFIDEV_BUSWIDTH_2 */
#ifdef CFIDEV_BUSWIDTH_4
	case CFIDEV_BUSWIDTH_4:
		switch (interleave) {
#ifdef CFIDEV_INTERLEAVE_1
		case CFIDEV_INTERLEAVE_1: /* 1 x32 device in x32 mode */
			val = cmd;
			break;
#endif
#ifdef CFIDEV_INTERLEAVE_2
		case CFIDEV_INTERLEAVE_2: /* 2 x16 device in x16 mode */
			val = (cmd << 16) | cmd;
			break;
#endif
#ifdef CFIDEV_INTERLEAVE_4
		case CFIDEV_INTERLEAVE_4: /* 4 (x8, x16 or x32) devices in x8 mode */
			val = (cmd << 16) | cmd;
			val = (val << 8) | val;
			break;
#endif
		}
		break;
#endif /* CFIDEV_BUSWIDTH_4 */
	}
	return val;
}

/*
 * Sends a CFI command to a bank of flash for the given geometry.
 *
 * Returns the offset in flash where the command was written.
 * If prev_val is non-null, it will be set to the value at the command address,
 * before the command was written.
 */
extern __inline__ __u32 cfi_send_gen_cmd(u_char cmd, __u32 cmd_addr, __u32 base, struct map_info *map,
					 u_char interleave, int type, __u32 *prev_val)
{
	__u32 val;
	__u32 addr = base + cfi_build_cmd_addr(cmd_addr, interleave, type);

	val = cfi_build_cmd(cmd, map, interleave);

	switch(map->buswidth) {
#ifdef CFIDEV_BUSWIDTH_1
	case CFIDEV_BUSWIDTH_1: /* 1 x8 device */
		if (prev_val) {
			*prev_val = map->read8(map, addr);
		}
		map->write8(map, val, addr);
		break;
#endif
#ifdef CFIDEV_BUSWIDTH_2
	case CFIDEV_BUSWIDTH_2:
		if (prev_val) {
			*prev_val = map->read16(map, addr);
		}
		map->write16(map, cpu_to_le16(val), addr);
		break;
#endif
#ifdef CFIDEV_BUSWIDTH_4
	case CFIDEV_BUSWIDTH_4:
		if (prev_val) {
			*prev_val = map->read32(map, addr);
		}
		map->write32(map, cpu_to_le32(val), addr);
		break;
	}
#endif
	return addr - base;
}


#endif /* __MTD_CFI_H__ */
